"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var UnifyClasses_1 = require("./UnifyClasses");
function flattenUnions(graph, stringTypeMapping, conflateNumbers, makeObjectTypes) {
    var needsRepeat = false;
    function replace(types, builder, forwardingRef) {
        var unionBuilder = new UnifyClasses_1.UnifyUnionBuilder(builder, true, makeObjectTypes, true, function (trefs, attributes) {
            Support_1.assert(trefs.length > 0, "Must have at least one type to build union");
            trefs = trefs.map(function (tref) { return builder.reconstituteType(tref.deref()[0]); });
            if (trefs.length === 1) {
                return trefs[0];
            }
            needsRepeat = true;
            return builder.getUnionType(attributes, immutable_1.OrderedSet(trefs));
        });
        return UnifyClasses_1.unifyTypes(types, immutable_1.Map(), builder, unionBuilder, conflateNumbers, forwardingRef);
    }
    var allUnions = graph.allTypesUnordered().filter(function (t) { return t instanceof Type_1.UnionType; });
    var nonCanonicalUnions = allUnions.filter(function (u) { return !u.isCanonical; });
    var foundIntersection = false;
    var groups = Type_1.makeGroupsToFlatten(nonCanonicalUnions, function (members) {
        Support_1.assert(!members.isEmpty(), "We can't have an empty union");
        if (!members.some(function (m) { return m instanceof Type_1.IntersectionType; }))
            return true;
        // FIXME: This is stupid.  `flattenUnions` returns true when no more union
        // flattening is necessary, but `resolveIntersections` can introduce new
        // unions that might require flattening, so now `flattenUnions` needs to take
        // that into account.  Either change `resolveIntersections` such that it
        // doesn't introduce non-canonical unions (by using `unifyTypes`), or have
        // some other way to tell whether more work is needed that doesn't require
        // the two passes to know about each other.
        foundIntersection = true;
        return false;
    });
    graph = graph.rewrite("flatten unions", stringTypeMapping, false, groups, replace);
    // console.log(`flattened ${nonCanonicalUnions.size} of ${unions.size} unions`);
    return [graph, !needsRepeat && !foundIntersection];
}
exports.flattenUnions = flattenUnions;
