"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function isValidJSON(s) {
    try {
        JSON.parse(s);
        return true;
    }
    catch (error) {
        return false;
    }
}
function sourcesFromPostmanCollection(collectionJSON) {
    var sources = [];
    var descriptions = [];
    function processCollection(c) {
        if (typeof c !== "object")
            return;
        if (Array.isArray(c.item)) {
            for (var _i = 0, _a = c.item; _i < _a.length; _i++) {
                var item = _a[_i];
                processCollection(item);
            }
            if (typeof c.info === "object" && typeof c.info.description === "string") {
                descriptions.push(c.info.description);
            }
        }
        if (typeof c.name === "string" && Array.isArray(c.response)) {
            var samples = [];
            for (var _b = 0, _c = c.response; _b < _c.length; _b++) {
                var r = _c[_b];
                if (typeof r === "object" && typeof r.body === "string" && isValidJSON(r.body)) {
                    samples.push(r.body);
                }
            }
            if (samples.length > 0) {
                var source = { kind: "json", name: c.name, samples: samples };
                var sourceDescription = [c.name];
                if (typeof c.request === "object") {
                    var _d = c.request, method = _d.method, url = _d.url;
                    if (method !== undefined && typeof url === "object" && url.raw !== undefined) {
                        sourceDescription.push(method + " " + url.raw);
                    }
                }
                if (typeof c.request === "object" && typeof c.request.description === "string") {
                    sourceDescription.push(c.request.description);
                }
                source.description = sourceDescription.length === 0 ? undefined : sourceDescription.join("\n\n");
                sources.push(source);
            }
        }
    }
    processCollection(JSON.parse(collectionJSON));
    var joinedDescription = descriptions.join("\n\n").trim();
    var description = undefined;
    if (joinedDescription !== "") {
        description = joinedDescription;
    }
    return { sources: sources, description: description };
}
exports.sourcesFromPostmanCollection = sourcesFromPostmanCollection;
