"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var TypeBuilder_1 = require("./TypeBuilder");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeAttributes_1 = require("./TypeAttributes");
function canResolve(t) {
    var members = Type_1.setOperationMembersRecursively(t)[0];
    if (members.size <= 1)
        return true;
    return members.every(function (m) { return !(m instanceof Type_1.UnionType) || m.isCanonical; });
}
function attributesForTypes(types) {
    return types
        .toMap()
        .map(function (t) { return t.getAttributes(); })
        .mapKeys(function (t) { return t.kind; });
}
var IntersectionAccumulator = /** @class */ (function () {
    function IntersectionAccumulator() {
        this._primitiveStringAttributes = immutable_1.OrderedMap();
        this._otherPrimitiveAttributes = immutable_1.OrderedMap();
        this._enumAttributes = TypeAttributes_1.emptyTypeAttributes;
        this._arrayAttributes = TypeAttributes_1.emptyTypeAttributes;
        // We start out with all object types allowed, which means
        // _additionalPropertyTypes is empty - no restrictions - and
        // _classProperties is empty - no defined properties so far.
        //
        // If _additionalPropertyTypes is undefined, no additional
        // properties are allowed anymore.  If _classProperties is
        // undefined, no object types are allowed, in which case
        // _additionalPropertyTypes must also be undefined;
        this._objectProperties = immutable_1.OrderedMap();
        this._objectAttributes = TypeAttributes_1.emptyTypeAttributes;
        this._additionalPropertyTypes = immutable_1.OrderedSet();
        this._lostTypeAttributes = false;
    }
    IntersectionAccumulator.prototype.updatePrimitiveStringTypes = function (members) {
        var types = members.filter(function (t) { return Type_1.isPrimitiveStringTypeKind(t.kind); });
        var attributes = attributesForTypes(types);
        this._primitiveStringAttributes = this._primitiveStringAttributes.mergeWith(TypeAttributes_1.combineTypeAttributes, attributes);
        var kinds = types.map(function (t) { return t.kind; });
        if (this._primitiveStringTypes === undefined) {
            this._primitiveStringTypes = kinds;
            return;
        }
        // If the unrestricted string type is part of the union, this doesn't add
        // any more restrictions.
        if (members.find(function (t) { return t instanceof Type_1.StringType; }) === undefined) {
            this._primitiveStringTypes = this._primitiveStringTypes.intersect(kinds);
        }
    };
    IntersectionAccumulator.prototype.updateOtherPrimitiveTypes = function (members) {
        var types = members.filter(function (t) { return Type_1.isPrimitiveTypeKind(t.kind) && !Type_1.isPrimitiveStringTypeKind(t.kind); });
        var attributes = attributesForTypes(types);
        this._otherPrimitiveAttributes = this._otherPrimitiveAttributes.mergeWith(TypeAttributes_1.combineTypeAttributes, attributes);
        var kinds = types.map(function (t) { return t.kind; });
        if (this._otherPrimitiveTypes === undefined) {
            this._otherPrimitiveTypes = kinds;
            return;
        }
        var haveNumber = this._otherPrimitiveTypes.find(Type_1.isNumberTypeKind) !== undefined &&
            kinds.find(Type_1.isNumberTypeKind) !== undefined;
        this._otherPrimitiveTypes = this._otherPrimitiveTypes.intersect(kinds);
        if (haveNumber && this._otherPrimitiveTypes.find(Type_1.isNumberTypeKind) === undefined) {
            // One set has integer, the other has double.  The intersection
            // of that is integer.
            this._otherPrimitiveTypes = this._otherPrimitiveTypes.add("integer");
        }
    };
    IntersectionAccumulator.prototype.updateEnumCases = function (members) {
        var enums = members.filter(function (t) { return t instanceof Type_1.EnumType; });
        var attributes = Type_1.combineTypeAttributesOfTypes(enums);
        this._enumAttributes = TypeAttributes_1.combineTypeAttributes(this._enumAttributes, attributes);
        if (members.find(function (t) { return t instanceof Type_1.StringType; }) !== undefined) {
            return;
        }
        var newCases = (_a = immutable_1.OrderedSet()).union.apply(_a, enums.map(function (t) { return t.cases; }).toArray());
        if (this._enumCases === undefined) {
            this._enumCases = newCases;
        }
        else {
            this._enumCases = this._enumCases.intersect(newCases);
        }
        var _a;
    };
    IntersectionAccumulator.prototype.updateArrayItemTypes = function (members) {
        var maybeArray = members.find(function (t) { return t instanceof Type_1.ArrayType; });
        if (maybeArray === undefined) {
            this._arrayItemTypes = false;
            return;
        }
        this._arrayAttributes = TypeAttributes_1.combineTypeAttributes(this._arrayAttributes, maybeArray.getAttributes());
        if (this._arrayItemTypes === undefined) {
            this._arrayItemTypes = immutable_1.OrderedSet();
        }
        else if (this._arrayItemTypes !== false) {
            this._arrayItemTypes = this._arrayItemTypes.add(maybeArray.items);
        }
    };
    IntersectionAccumulator.prototype.updateObjectProperties = function (members) {
        var _this = this;
        var maybeObject = members.find(function (t) { return t instanceof Type_1.ObjectType; });
        if (maybeObject === undefined) {
            this._objectProperties = undefined;
            this._additionalPropertyTypes = undefined;
            return;
        }
        this._objectAttributes = TypeAttributes_1.combineTypeAttributes(this._objectAttributes, maybeObject.getAttributes());
        if (this._objectProperties === undefined) {
            Support_1.assert(this._additionalPropertyTypes === undefined);
            return;
        }
        var allPropertyNames = this._objectProperties.keySeq().toOrderedSet().union(maybeObject.properties.keySeq());
        allPropertyNames.forEach(function (name) {
            var existing = Support_1.defined(_this._objectProperties).get(name);
            var newProperty = maybeObject.properties.get(name);
            if (existing !== undefined && newProperty !== undefined) {
                var cp = new Type_1.GenericClassProperty(existing.typeData.add(newProperty.type), existing.isOptional || newProperty.isOptional);
                _this._objectProperties = Support_1.defined(_this._objectProperties).set(name, cp);
            }
            else if (existing !== undefined && maybeObject.additionalProperties !== undefined) {
                var cp = new Type_1.GenericClassProperty(existing.typeData.add(maybeObject.additionalProperties), existing.isOptional);
                _this._objectProperties = Support_1.defined(_this._objectProperties).set(name, cp);
            }
            else if (existing !== undefined) {
                _this._objectProperties = Support_1.defined(_this._objectProperties).remove(name);
            }
            else if (newProperty !== undefined && _this._additionalPropertyTypes !== undefined) {
                var types = _this._additionalPropertyTypes.add(newProperty.type);
                _this._objectProperties = Support_1.defined(_this._objectProperties).set(name, new Type_1.GenericClassProperty(types, newProperty.isOptional));
            }
            else if (newProperty !== undefined) {
                _this._objectProperties = Support_1.defined(_this._objectProperties).remove(name);
            }
            else {
                return Support_1.panic("This should not happen");
            }
        });
        if (this._additionalPropertyTypes !== undefined && maybeObject.additionalProperties) {
            this._additionalPropertyTypes = this._additionalPropertyTypes.add(maybeObject.additionalProperties);
        }
        else if (this._additionalPropertyTypes !== undefined || maybeObject.additionalProperties) {
            this._additionalPropertyTypes = undefined;
            this._lostTypeAttributes = true;
        }
    };
    IntersectionAccumulator.prototype.addUnionSet = function (members) {
        this.updatePrimitiveStringTypes(members);
        this.updateOtherPrimitiveTypes(members);
        this.updateEnumCases(members);
        this.updateArrayItemTypes(members);
        this.updateObjectProperties(members);
    };
    IntersectionAccumulator.prototype.addType = function (t) {
        var _this = this;
        var attributes = t.getAttributes();
        Type_1.matchTypeExhaustive(t, function (_noneType) {
            return Support_1.panic("There shouldn't be a none type");
        }, function (_anyType) {
            return Support_1.panic("The any type should have been filtered out in setOperationMembersRecursively");
        }, function (nullType) { return _this.addUnionSet(immutable_1.OrderedSet([nullType])); }, function (boolType) { return _this.addUnionSet(immutable_1.OrderedSet([boolType])); }, function (integerType) { return _this.addUnionSet(immutable_1.OrderedSet([integerType])); }, function (doubleType) { return _this.addUnionSet(immutable_1.OrderedSet([doubleType])); }, function (stringType) { return _this.addUnionSet(immutable_1.OrderedSet([stringType])); }, function (arrayType) { return _this.addUnionSet(immutable_1.OrderedSet([arrayType])); }, function (_classType) { return Support_1.panic("We should never see class types in intersections"); }, function (_mapType) { return Support_1.panic("We should never see map types in intersections"); }, function (objectType) { return _this.addUnionSet(immutable_1.OrderedSet([objectType])); }, function (enumType) { return _this.addUnionSet(immutable_1.OrderedSet([enumType])); }, function (unionType) {
            attributes = TypeAttributes_1.combineTypeAttributes([attributes].concat(unionType.members.toArray().map(function (m) { return m.getAttributes(); })));
            _this.addUnionSet(unionType.members);
        }, function (dateType) { return _this.addUnionSet(immutable_1.OrderedSet([dateType])); }, function (timeType) { return _this.addUnionSet(immutable_1.OrderedSet([timeType])); }, function (dateTimeType) { return _this.addUnionSet(immutable_1.OrderedSet([dateTimeType])); });
        return TypeAttributes_1.makeTypeAttributesInferred(attributes);
    };
    Object.defineProperty(IntersectionAccumulator.prototype, "arrayData", {
        get: function () {
            if (this._arrayItemTypes === undefined || this._arrayItemTypes === false) {
                return Support_1.panic("This should not be called if the type can't be an array");
            }
            return this._arrayItemTypes;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "objectData", {
        get: function () {
            if (this._objectProperties === undefined) {
                Support_1.assert(this._additionalPropertyTypes === undefined);
                return undefined;
            }
            return [this._objectProperties, this._additionalPropertyTypes];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "enumCases", {
        get: function () {
            return Support_1.defined(this._enumCases).toArray();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(IntersectionAccumulator.prototype, "enumCaseMap", {
        get: function () {
            var caseMap = {};
            Support_1.defined(this._enumCases).forEach(function (n) { return (caseMap[n] = 1); });
            return caseMap;
        },
        enumerable: true,
        configurable: true
    });
    IntersectionAccumulator.prototype.getMemberKinds = function () {
        var _this = this;
        var primitiveStringKinds = Support_1.defined(this._primitiveStringTypes)
            .toOrderedMap()
            .map(function (k) { return Support_1.defined(_this._primitiveStringAttributes.get(k)); });
        var maybeStringAttributes = this._primitiveStringAttributes.get("string");
        // If full string was eliminated, add its attribute to the other string types
        if (maybeStringAttributes !== undefined && !primitiveStringKinds.has("string")) {
            primitiveStringKinds = primitiveStringKinds.map(function (a) { return TypeAttributes_1.combineTypeAttributes(a, maybeStringAttributes); });
        }
        var otherPrimitiveKinds = Support_1.defined(this._otherPrimitiveTypes)
            .toOrderedMap()
            .map(function (k) { return Support_1.defined(_this._otherPrimitiveAttributes.get(k)); });
        var maybeDoubleAttributes = this._otherPrimitiveAttributes.get("double");
        // If double was eliminated, add its attributes to integer
        if (maybeDoubleAttributes !== undefined && !otherPrimitiveKinds.has("double")) {
            otherPrimitiveKinds = otherPrimitiveKinds.map(function (a, k) {
                if (k !== "integer")
                    return a;
                return TypeAttributes_1.combineTypeAttributes(a, maybeDoubleAttributes);
            });
        }
        var kinds = primitiveStringKinds.merge(otherPrimitiveKinds);
        if (this._enumCases !== undefined && this._enumCases.size > 0) {
            kinds = kinds.set("enum", this._enumAttributes);
        }
        else if (!this._enumAttributes.isEmpty()) {
            if (kinds.has("string")) {
                kinds = kinds.update("string", function (ta) { return TypeAttributes_1.combineTypeAttributes(ta, _this._enumAttributes); });
            }
            else {
                this._lostTypeAttributes = true;
            }
        }
        if (immutable_1.OrderedSet.isOrderedSet(this._arrayItemTypes)) {
            kinds = kinds.set("array", this._arrayAttributes);
        }
        else if (!this._arrayAttributes.isEmpty()) {
            this._lostTypeAttributes = true;
        }
        if (this._objectProperties !== undefined) {
            kinds = kinds.set("object", this._objectAttributes);
        }
        else if (!this._objectAttributes.isEmpty()) {
            this._lostTypeAttributes = true;
        }
        return kinds;
    };
    Object.defineProperty(IntersectionAccumulator.prototype, "lostTypeAttributes", {
        get: function () {
            return this._lostTypeAttributes;
        },
        enumerable: true,
        configurable: true
    });
    return IntersectionAccumulator;
}());
var IntersectionUnionBuilder = /** @class */ (function (_super) {
    __extends(IntersectionUnionBuilder, _super);
    function IntersectionUnionBuilder() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._createdNewIntersections = false;
        return _this;
    }
    IntersectionUnionBuilder.prototype.makeIntersection = function (members, attributes) {
        var _this = this;
        var reconstitutedMembers = members.map(function (t) { return _this.typeBuilder.reconstituteTypeRef(t.typeRef); });
        var first = Support_1.defined(reconstitutedMembers.first());
        if (reconstitutedMembers.size === 1) {
            this.typeBuilder.addAttributes(first, attributes);
            return first;
        }
        this._createdNewIntersections = true;
        return this.typeBuilder.getUniqueIntersectionType(attributes, reconstitutedMembers);
    };
    Object.defineProperty(IntersectionUnionBuilder.prototype, "createdNewIntersections", {
        get: function () {
            return this._createdNewIntersections;
        },
        enumerable: true,
        configurable: true
    });
    IntersectionUnionBuilder.prototype.makeEnum = function (cases, _counts, typeAttributes, forwardingRef) {
        return this.typeBuilder.getEnumType(typeAttributes, immutable_1.OrderedSet(cases), forwardingRef);
    };
    IntersectionUnionBuilder.prototype.makeObject = function (maybeData, typeAttributes, forwardingRef) {
        var _this = this;
        if (maybeData === undefined) {
            return Support_1.panic("Either properties or additional properties must be given to make an object type");
        }
        var propertyTypes = maybeData[0], maybeAdditionalProperties = maybeData[1];
        var properties = propertyTypes.map(function (cp) { return new Type_1.ClassProperty(_this.makeIntersection(cp.typeData, immutable_1.Map()), cp.isOptional); });
        var additionalProperties = maybeAdditionalProperties === undefined
            ? undefined
            : this.makeIntersection(maybeAdditionalProperties, TypeAttributes_1.emptyTypeAttributes);
        return this.typeBuilder.getUniqueObjectType(typeAttributes, properties, additionalProperties, forwardingRef);
    };
    IntersectionUnionBuilder.prototype.makeArray = function (arrays, typeAttributes, forwardingRef) {
        // FIXME: attributes
        var itemsType = this.makeIntersection(arrays, immutable_1.Map());
        var tref = this.typeBuilder.getArrayType(itemsType, forwardingRef);
        this.typeBuilder.addAttributes(tref, typeAttributes);
        return tref;
    };
    return IntersectionUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function resolveIntersections(graph, stringTypeMapping) {
    var needsRepeat = false;
    function replace(types, builder, forwardingRef) {
        var intersections = types.filter(function (t) { return t instanceof Type_1.IntersectionType; });
        var _a = Type_1.setOperationMembersRecursively(intersections.toArray()), members = _a[0], intersectionAttributes = _a[1];
        if (members.isEmpty()) {
            var t = builder.getPrimitiveType("any", forwardingRef);
            builder.addAttributes(t, intersectionAttributes);
            return t;
        }
        if (members.size === 1) {
            var single = builder.reconstituteType(Support_1.defined(members.first()), forwardingRef);
            builder.addAttributes(single, intersectionAttributes);
            return single;
        }
        var accumulator = new IntersectionAccumulator();
        var extraAttributes = TypeAttributes_1.makeTypeAttributesInferred(TypeAttributes_1.combineTypeAttributes(members.toArray().map(function (t) { return accumulator.addType(t); })));
        var attributes = TypeAttributes_1.combineTypeAttributes(intersectionAttributes, extraAttributes);
        var unionBuilder = new IntersectionUnionBuilder(builder);
        var tref = unionBuilder.buildUnion(accumulator, true, attributes, forwardingRef);
        if (unionBuilder.createdNewIntersections) {
            needsRepeat = true;
        }
        return tref;
    }
    // FIXME: We need to handle intersections that resolve to the same set of types.
    // See for example the intersections-nested.schema example.
    var allIntersections = graph.allTypesUnordered().filter(function (t) { return t instanceof Type_1.IntersectionType; });
    var resolvableIntersections = allIntersections.filter(canResolve);
    var groups = Type_1.makeGroupsToFlatten(resolvableIntersections, undefined);
    graph = graph.rewrite("resolve intersections", stringTypeMapping, false, groups, replace);
    // console.log(`resolved ${resolvableIntersections.size} of ${intersections.size} intersections`);
    return [graph, !needsRepeat && allIntersections.size === resolvableIntersections.size];
}
exports.resolveIntersections = resolveIntersections;
