"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Support_1 = require("./Support");
var TypeAttributeKind = /** @class */ (function () {
    function TypeAttributeKind(name, combine, makeInferred, stringify) {
        this.name = name;
        if (combine === undefined) {
            combine = function () {
                return Support_1.panic("Cannot combine type attribute " + name);
            };
        }
        this.combine = combine;
        if (makeInferred === undefined) {
            makeInferred = function () {
                return Support_1.panic("Cannot make type attribute " + name + " inferred");
            };
        }
        this.makeInferred = makeInferred;
        if (stringify === undefined) {
            stringify = function () { return undefined; };
        }
        this.stringify = stringify;
    }
    TypeAttributeKind.prototype.makeAttributes = function (value) {
        var kvps = [[this, value]];
        return immutable_1.Map(kvps);
    };
    TypeAttributeKind.prototype.tryGetInAttributes = function (a) {
        return a.get(this);
    };
    TypeAttributeKind.prototype.setInAttributes = function (a, value) {
        return a.set(this, value);
    };
    TypeAttributeKind.prototype.modifyInAttributes = function (a, modify) {
        var modified = modify(this.tryGetInAttributes(a));
        if (modified === undefined) {
            return a.remove(this);
        }
        return this.setInAttributes(a, modified);
    };
    TypeAttributeKind.prototype.setDefaultInAttributes = function (a, makeDefault) {
        if (this.tryGetInAttributes(a) !== undefined)
            return a;
        return this.modifyInAttributes(a, makeDefault);
    };
    TypeAttributeKind.prototype.equals = function (other) {
        if (!(other instanceof TypeAttributeKind)) {
            return false;
        }
        return this.name === other.name;
    };
    TypeAttributeKind.prototype.hashCode = function () {
        return immutable_1.hash(this.name);
    };
    return TypeAttributeKind;
}());
exports.TypeAttributeKind = TypeAttributeKind;
exports.emptyTypeAttributes = immutable_1.Map();
function combineTypeAttributes(firstOrArray, second) {
    var attributeArray;
    var first;
    var rest;
    if (Array.isArray(firstOrArray)) {
        attributeArray = firstOrArray;
        if (attributeArray.length === 0)
            return immutable_1.Map();
        first = attributeArray[0];
        rest = attributeArray.slice(1);
    }
    else {
        if (second === undefined) {
            return Support_1.panic("Must have on array or two attributes");
        }
        first = firstOrArray;
        rest = [second];
    }
    return first.mergeWith.apply(first, [function (aa, ab, kind) { return kind.combine(aa, ab); }].concat(rest));
}
exports.combineTypeAttributes = combineTypeAttributes;
function makeTypeAttributesInferred(attr) {
    return attr.map(function (value, kind) { return kind.makeInferred(value); }).filter(function (v) { return v !== undefined; });
}
exports.makeTypeAttributesInferred = makeTypeAttributesInferred;
exports.descriptionTypeAttributeKind = new TypeAttributeKind("description", Support_1.setUnion, function (_) { return immutable_1.OrderedSet(); }, undefined);
exports.propertyDescriptionsTypeAttributeKind = new TypeAttributeKind("propertyDescriptions", function (a, b) { return a.mergeWith(Support_1.setUnion, b); }, function (_) { return immutable_1.Map(); }, undefined);
