"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var typescript_json_schema_1 = require("typescript-json-schema");
var Support_1 = require("./Support");
var EncodedDefaultTypeScriptLibrary_1 = require("./EncodedDefaultTypeScriptLibrary");
var settings = {
    required: true,
    titles: true,
    topRef: true
};
var compilerOptions = {
    strictNullChecks: true,
    typeRoots: []
};
var libFileName = "lib.d.ts";
var libSource = undefined;
function getLibSource() {
    if (libSource === undefined) {
        libSource = Support_1.inflateBase64(EncodedDefaultTypeScriptLibrary_1.encodedDefaultTypeScriptLibrary);
    }
    return libSource;
}
var CompilerHost = /** @class */ (function () {
    function CompilerHost(_options, _sources) {
        this._sources = _sources;
    }
    CompilerHost.prototype.fileExists = function (fileName) {
        if (fileName === libFileName)
            return true;
        return Object.prototype.hasOwnProperty.apply(this._sources, fileName);
    };
    CompilerHost.prototype.readFile = function (fileName) {
        if (fileName === libFileName) {
            return getLibSource();
        }
        return this._sources[fileName];
    };
    CompilerHost.prototype.getSourceFile = function (fileName, languageVersion, _onError, _shouldCreateNewSourceFile) {
        var sourceText = this.readFile(fileName);
        return sourceText !== undefined ? ts.createSourceFile(fileName, sourceText, languageVersion) : undefined;
    };
    CompilerHost.prototype.getDefaultLibFileName = function (_options) {
        return libFileName;
    };
    CompilerHost.prototype.writeFile = function (_fileName, _data, _writeByteOrderMark, _onError, _sourceFiles) {
        return Support_1.panic("cannot write file");
    };
    CompilerHost.prototype.getCurrentDirectory = function () {
        return ".";
    };
    CompilerHost.prototype.getDirectories = function (_path) {
        return [];
    };
    CompilerHost.prototype.getCanonicalFileName = function (fileName) {
        if (this.useCaseSensitiveFileNames()) {
            return fileName.toLowerCase();
        }
        return fileName;
    };
    CompilerHost.prototype.useCaseSensitiveFileNames = function () {
        return false;
    };
    CompilerHost.prototype.getNewLine = function () {
        return "\n";
    };
    return CompilerHost;
}());
function makeCompilerOptions(jsonCompilerOptions, basePath) {
    if (basePath === void 0) { basePath = "./"; }
    var compilerOptions = ts.convertCompilerOptionsFromJson(jsonCompilerOptions, basePath).options;
    var options = {
        noEmit: true, emitDecoratorMetadata: true, experimentalDecorators: true, target: ts.ScriptTarget.ES5, module: ts.ModuleKind.CommonJS
    };
    for (var k in compilerOptions) {
        if (compilerOptions.hasOwnProperty(k)) {
            options[k] = compilerOptions[k];
        }
    }
    return options;
}
function schemaForTypeScriptSources(sources) {
    var options = makeCompilerOptions(compilerOptions);
    var fileNames;
    var host;
    if (Array.isArray(sources)) {
        fileNames = sources;
        host = ts.createCompilerHost(options);
    }
    else {
        fileNames = Object.getOwnPropertyNames(sources);
        host = new CompilerHost(options, sources);
    }
    var program = ts.createProgram(fileNames, options, host);
    var schema = typescript_json_schema_1.generateSchema(program, "*", settings);
    return JSON.stringify(schema);
}
exports.schemaForTypeScriptSources = schemaForTypeScriptSources;
