/// <reference types="node" />
import { Map, OrderedMap } from "immutable";
import { Readable } from "stream";
import { TargetLanguage } from "./TargetLanguage";
import { SerializedRenderResult } from "./Source";
import { JSONSchemaStore } from "./JSONSchemaStore";
export { TargetLanguage } from "./TargetLanguage";
export { SerializedRenderResult, Annotation } from "./Source";
export { all as languages, languageNamed } from "./Language/All";
export { OptionDefinition } from "./RendererOptions";
export declare function getTargetLanguage(nameOrInstance: string | TargetLanguage): TargetLanguage;
export declare type RendererOptions = {
    [name: string]: string;
};
export declare type StringInput = string | Readable;
export interface JSONTypeSource {
    kind: "json";
    name: string;
    samples: StringInput[];
    description?: string;
}
export interface TypeScriptTypeSource {
    kind: "typescript";
    sources: {
        [filename: string]: string;
    };
}
export declare function isTypeScriptSource(source: TypeSource): source is TypeScriptTypeSource;
export interface SchemaTypeSource {
    kind: "schema";
    name: string;
    uri?: string;
    schema?: StringInput;
    topLevelRefs?: string[];
}
export interface GraphQLTypeSource {
    kind: "graphql";
    name: string;
    schema: any;
    query: StringInput;
}
export declare type TypeSource = GraphQLTypeSource | JSONTypeSource | SchemaTypeSource | TypeScriptTypeSource;
export interface Options {
    lang: string | TargetLanguage;
    sources: TypeSource[];
    handlebarsTemplate: string | undefined;
    findSimilarClassesSchemaURI: string | undefined;
    inferMaps: boolean;
    inferEnums: boolean;
    inferDates: boolean;
    alphabetizeProperties: boolean;
    allPropertiesOptional: boolean;
    combineClasses: boolean;
    fixedTopLevels: boolean;
    noRender: boolean;
    leadingComments: string[] | undefined;
    rendererOptions: RendererOptions;
    indentation: string | undefined;
    outputFilename: string;
    schemaStore: JSONSchemaStore | undefined;
    debugPrintGraph: boolean;
    checkProvenance: boolean;
}
export declare class Run {
    private _compressedJSON;
    private _allInputs;
    private _options;
    private _schemaStore;
    constructor(options: Partial<Options>);
    private getSchemaStore();
    private makeGraph();
    private makeSimpleTextResult(lines);
    private addSchemaInput(name, ref);
    run: () => Promise<OrderedMap<string, SerializedRenderResult>>;
}
export declare function quicktypeMultiFile(options: Partial<Options>): Promise<Map<string, SerializedRenderResult>>;
export declare function quicktype(options: Partial<Options>): Promise<SerializedRenderResult>;
